---
jupyter:
  jupytext:
    formats: ipynb,md
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.1'
      jupytext_version: 1.1.1
  kernelspec:
    display_name: pycroc
    language: python
    name: pycroc
---

# Tutorial
********



## Summary

Powerful biomarkers are important tools in diagnostic, clinical and research settings. In the area of diagnostic medicine, a biomarker is often used as a tool to identify subjects with a disease, or at high risk of developing a disease. Moreover, it can be used to foresee the more likely outcome of the disease, monitor its progression and predict the response to a given therapy. Diagnostic accuracy can be improved considerably by combining multiple markers, whose performance in identifying diseased subjects is usually assessed via receiver operating characteristic (ROC) curves. The CombiROC tool was originally designed as an easy to use R-Shiny web application to determine optimal combinations of markers from diverse complex omics data ([Mazzara et al. 2017](https://www.nature.com/articles/srep45477)); such an implementation is easy to use but has limited features and limitations arise from the machine it is deployed on. The CombiROC package is the natural evolution of the CombiROC tool and it allows the researcher/analyst to freely use the method and further build on it.



## The complete workflow

This python package was developed as porting for the popular CombiROC workflow. All the code and tutorial were developed with this work in mind. 

The aim of this document is to show the whole CombiROC workflow to get you up and running as quickly as possible with this package. To do so we’re going to use the proteomic dataset from Zingaretti et al. 2012 containing multi-marker signatures for Autoimmune Hepatitis (AIH) for samples clinically diagnosed as “abnormal” (class A) or “normal” (class B). The scope of the workflow is to first find the markers combinations, then to assess their performance in classifying samples of the dataset.

Note: if you use CombiROC in your research, please cite:

> [Mazzara S., Rossi R.L., Grifantini R., Donizetti S., Abrignani L., Bombaci M. (2017) CombiROC: an interactive web tool for selecting accurate marker combinations of omics data. Scientific Reports, 7:45477. 10.1038/srep45477](https://www.nature.com/articles/srep45477)

> [Ferrari I., Mazzara S., Abrignani S., Grifantini R.M., Bombaci M., Rossi R.L. (2022) Combinatorial selection of biomarkers to optimize gene signatures in diagnostics and single cell applications. bioRxiv, 10.1101/2022.01.17.476603 ](https://www.biorxiv.org/content/10.1101/2022.01.17.476603v1)




## Required data format

The dataset to be analysed should be in text format, which can be separated by commas, tabs or semicolons. Format of the columns should be the following:

- The 1st column must contain unique patient/sample IDs.
- The 2nd column must contain the class to which each sample belongs.
- The classes must be exactly TWO and they must be labelled with character format with “A” (usually the cases) and “B” (usually the controls).
- From the 3rd column on, the dataset must contain numerical values that represent the signal corresponding to the markers abundance in each sample (marker-related columns).
- The header for marker-related columns can be called ‘Marker1, Marker2, Marker3, …’ or can be called directly with the gene/protein name. Please note that “-” (dash) is not allowed in the column names



## Data loading

The `load_data` function uses a customized `read.table` function that checks the conformity of the dataset format. If all the checks are passed, marker-related columns are reordered alphabetically, depending on marker names (this is necessary for a proper computation of combinations), and it imposes _Class_ as the name of the second column. The loaded dataset is here assigned to the _data_ object. 

Please note that `load_data` takes the semicolumn (“;”) as default separator: if the dataset to be loaded has a different separator, i.e. a comma (“,”), is necessary to specify it in the argument sep.

The code below shows how to load a data set contained in the “data” folder (remember to adjust the path according to your current working directory):


```python
import os
import pandas as pd
import numpy as np
import itertools

import matplotlib
from matplotlib import colors
from matplotlib.colorlss import ListedColormap, LinearSegmentedColormap
from matplotlib import pyplot as plt
import plotly.figure_factory as ff
import seaborn as sns 

import statsmodels.api as sm # 0.9.0
import statsmodels.formula.api as smf
import sklearn
from sklearn import metrics, preprocessing

from scipy import stats

import scanpy as sc

import os
```

```python
import pycroc as pcr
```

```python
#data = pcr.load_data('./data/demo_data.csv', sep="\t")
```

For follow the tutorial you can load the example dataset as:

```python
data = pcr.datasets.Zingaretti_2012()
```

Look at the dataframe structure:

```python
data.head()
```

## Data exploration

It is usually a good thing to visually explore your data with at least few plots. To do this, pycroc provides the class `Distr` that initialize an object directly from the data after the case-class specification


```python
distr = pcr.Distr(data, case_class='A')
```

Box plots are a nice option to observe the distribution of measurements in each sample. The user can plot the data as she/he wishes using the preferred function.



```python
distr.markers_boxplot(ylim=(0,2000))
```

## The ROC curve for all markers and its coordinates

The [ROC curve](https://it.wikipedia.org/wiki/Receiver_operating_characteristic) shows how many real positive samples would be found positive (sensitivity, or SE) and how many real negative samples would be found negative (specificity, or SP) in function of signal threshold. Please note that the False Positive Rate (i.e. 1 - specificity) is plotted on the x-axis. These SE and SP are refereed to the signal intensity threshold considering all the markers together; they are not the SE and SP of a single marker/combination computed by the se_sp() function further discussed in the Sensitivity and specificity paragraph.


```python
#initialize minimal values for SE and SP
distr.markers_coord(min_SE=40, min_SP=80)
print(distr.min_SE, distr.min_SP)
```

```python
distr.markers_ROC()
```

Users that want more control over the plotting process could access a long-format dataframe accessing the object attribute `data_long`

```python
distr.data_long.head()
```

The same is true fot hte resulting computed threshold usign the attribute `coord`

```python
distr.coord.head()
```

## The density plot and suggested signal threshold

The `markers_density` method shows the distribution of the signal intensity values for both the classes. 

In addition, the function allows the user to set both the ylim and xlim values to customize the visualization. 

One important feature of the density plot is that it calculates a possible signal intensity threshold: in case of lack of a priori knowedge of the threshold the user can set the argument `signalthr_prediction = TRUE`.

In this way the function calculates a “suggested signal threshold” that corresponds to the median of the signal threshold values (in Coord) at which SE and SP are greater or equal to their set minimal values (min_SE and min_SP). This threshold is added to the plot as a dashed black line and a number. The use of the median allows to pick a threshold whose SE/SP are not too close to the limits (min_SE and min_SP), but it is recommended to always inspect “Coord” and choose the most appropriate signal threshold by considering SP, SE and Youden index.

This suggested signal threshold can be used as signalthr argument of the combi() function further in the workflow.


```python
#     def markers_density(self, xlim=(0,None), ylim=(0, None), 
#                         bw_adjust= {'case':'nrd0', 'control':'nrd0'},
#                         colors= [(1.0, 0.4980392156862745, 0.054901960784313725), 
#                                  (0.12156862745098039, 0.4666666666666667, 0.7058823529411765)],
#                         signalthr_prediction=False):
  
#         #NOTE: if min_SE/min_SP were set before, compute threshold using sub_coord like object
#         #sns.set_palette(colors)

#         case= np.array(self.data_long.loc[self.data_long.Class==self.case_class,].Values)
#         control = np.array(self.data_long.loc[self.data_long.Class==self.control_class,].Values)

#         data={'case':case, 'control':control}

#         for i in ['case', 'control']:
#             if not i in bw_adjust.keys():
#                 bw_adjust[i]='nrd0'
#             if bw_adjust[i]=='nrd0':
#                 bw_adjust[i]= _bw_nrd0(data[i])
#             kde=sns.kdeplot(data[i],  bw_adjust=bw_adjust[i], gridsize=10000)

#         kde.set(ylim=ylim, xlim=xlim)
#         kde.legend([self.case_class,self.control_class])

#         if signalthr_prediction==True:
#             sub_coord = self.coord.loc[(self.coord['SE'] >= self.min_SE) & (self.coord['SP'] >= self.min_SP),:]
#             self.optimal_threshold = sub_coord.threshold.median()
#             kde.axvline(self.optimal_threshold, color='black', ls='--')
#             kde.text(self.optimal_threshold*0.9, 0.00002, str(round(self.optimal_threshold,2)), 
#                      horizontalalignment='right', size='small', color='black')
```

```python
distr.markers_density(ylim=(0,0.002), xlim=(0,4000), signalthr_prediction=True)
```

## Combinatorial analysis

Pycroc has a second class named `Combinations` used to compute the marker combinations and counts positive samples for each class (once thresholds are selected). 

A sample, to be considered positive for a given combination, must have a value higher than a given signal threshold (`signalthr`) for at least a given number of markers composing that combination (`combithr`).

As mentioned before, `signalthr` parameter should be set depending on the guidelines and characteristics of the methodology used for the assay or by an accurate inspection of signal intensity distribution. In case of lack of specific guidelines, one should set the value `signalthr` as suggested by the `markers_density` method, described in the previous section.

In this tutorial, we set `signalthr` equal to 450 instead of 407.5 in order to reproduce the results reported in Mazzara et. al 2017 (the original CombiROC paper) or in Bombaci & Rossi 2019 as well as in the tutorial of the web app with default thresholds.

`combithr`, instead, should be set exclusively depending on the needed stringency: 1 is the less stringent and most common choice (meaning that at least one marker in a combination needs to reach the threshold). The obtained tab object is a dataframe of all the combinations obtained with the chosen parameters.


```python
combi = pcr.Combinations(data, case_class='A',signalthr=450, combithr=1)
```

```python
combi.tab.head()
```

## Selection of combinations

The markers combinations can now be ranked and selected. After specifying the case class (“A” in this case), the method `ranked_combs` ranks the combinations by the [Youden index](https://en.wikipedia.org/wiki/Youden%27s_J_statistic) in order to show the combinations with the highest SE (of cases) and SP (of controls) on the top, facilitating the user in the selection of the best ones. Again, the Youden index (J) is calculated in this way:

<p style="text-align:center">$${J=SE+SP−1}$$</p>

The user can also set (not mandatory) a minimal value of SE and/or SP that a combination must have to be selected, i.e. to be considered as “gold” combinations.


```python
combi.ranked_combs()
combi.ranked
```

A possibility to overview how single markers and all combinations are distributed in the SE - SP ballpark is to plot them with the `bubble_chart` method.
The bigger the bubble, the more markers are in the combination: looking at the size and distribution of bubbles across SE and SP values is useful to anticipate how effective will be the combinations in the ranking. Setting no cutoffs (i.e. SE = 0 and SP = 0), all single markers and combinations (all bubbles) will be considered as `gold` combinations and ranked in the next passage.

In the the example below the minimal values of SE and SP are set, respectively, to 40 and 80, in order to reproduce the gold combinations selection reported in Mazzara et. al 2017. The obtained values of combinations, ranked according to Youden index, are stored in the `ranked` attribute.


```python
combi.bubble_chart(min_SE=40, min_SP=80)
```

## ROC curves

To allow an objective comparison of combinations, pycroc applies the Generalised Linear Model for each selected combination.


```python
#data for combination were alreadys tored in the combi object
combi.train_models(combinations=['Marker1', 'Combination 11', 'Combination 15', 'Combination 22'])
```

Model results were stored in the `models` attribute of the `combi` object.

```python
combi.models
```

The resulting predictions are then used to compute ROC curves (with function pROC::roc()) and their corresponding metrics which are both returned by the function as a named list object (in this case called reports). The function roc_reports() requires as input:

- The data object ( data ) obtained with load_data();
- The table with combinations and corresponding positive samples counts ( tab ), obtained with combi().

In addition, the user has to specify the class case, and the single markers and/or the combinations that she/he wants to be displayed with the specific function’s arguments.
In the example below a single marker ( Marker1 ) and two combinations (combinations number 11 and 15 ) were choosen.


```python
#combi.models['Combination 11'].predict() # combi score  a.k.a p(x)
```

```python
combi.roc_curves(case_class='A')
```

```python
combi.models['Combination 11'].summary()
```

```python
combi.reports
```

```python
pcr.combi_score(data, combi.models)
```

```python
pcr.combi_score(data, combi.models, classify=True, reports=combi.reports, case_class='A', control_class='B')
```

## Single-cell applications

```python
import scanpy as sc
```

```python
adata=sc.datasets.pbmc68k_reduced()
```

```python
adata
```

```python
gene_list=adata.uns['rank_genes_groups']['names']['CD56+ NK'][0:29]
```

```python
scanpy_to_pycroc(AnnData=adata, gene_list=gene_list, obs_name='bulk_labels', case_class=['CD56+ NK'], case_label='NK')
```

```python

```
